import numpy as np
from typing import List, Tuple
import pandas as pd


def correlation_matrix(m: np.ndarray, labels: List[str] = None) -> np.ndarray:
    """
        Calculates a correlation matrix given a :math:`[N_{samples} \\times N_{features}]` data matrix.
        
        This correlation is composed of the Pearson Correlation Coefficient
        between each pair of features.

        Args:
            m (:class:`numpy.ndarray`): 
                A :math:`[N_{samples} \\times N_{features}]` data matrix.
            labels (list(str)): 
                A list of n_features labels.

        Returns:
            :class:`numpy.ndarray`: The resulting :math:`[N_{features} \\times N_{features}]` correlation matrix.
    """
    corr_matrix = np.corrcoef(m)

    return corr_matrix


def distance_matrix(m: np.ndarray) -> np.ndarray:
    """
        Calculates a :math:`[N_{samples} \\times N_{samples}]` distance matrix given a :math:`[N_{samples} \\times N_{features}]` data matrix. 

        Args:
            m (:class:`numpy.ndarray`): A :math:`[N_{samples} \\times N_{features}]` data matrix.

        Returns:
            :class:`numpy.ndarray`: The resulting :math:`[N_{samples} \\times N_{samples}]` distance matrix.
    """
    from scipy.spatial.distance import pdist, squareform

    dist_matrix = squareform(pdist(m))

    return dist_matrix


def distance_sorter(dist_matrix: np.ndarray) -> Tuple[np.array, np.array]:
    """
        Calculates an sorted total distance (useful for finding outliers) and plots it.
        
        Args:
            dist_matrix (:class:`numpy.ndarray`): A [N x N] distance matrix.
        
        Returns:
            (tuple): Tuple contaning:
                - dists (:class:`numpy.ndarray`): Total distance for each sample.
                - i_dists (:class:`numpy.ndarray`): Indexes used to sort dists.
    """
    dists = np.sum(dist_matrix, axis=1)

    i_dists = np.argsort(dists)
    dists_sorted = dists[i_dists]

    return i_dists, dists_sorted


def count_unique_votes(df: pd.DataFrame) -> pd.DataFrame:
    """
        Counts amount of each vote for all products.

        Args:
            df (:class:`pandas.Dataframe`): The voting df with all votes for each product.
        Returns:
            :class:`pandas.Dataframe`: The original dataframe with additional columns for each note
            and their count
    """

    # Checks for the necessary columns in the dataframe
    if "id_produto_cor" not in df.columns:
        print("Dataframe does not contain id_produto_cor key column")
        return -1
    if "nota" not in df.colums:
        print("Dataframe does not contain id_produto_cor key column")
        return -2

    # Finding all different notes
    notes_list = df["nota"].unique().apply(lambda x: "C" + str(x)).tolist()

    d = {}
    for id in df["id_produto_cor"].unique():
        count = df[df["id_produto_cor"] == id]["nota"].value_counts(sort=False).T
        d[id] = count.to_numpy()

    return (
        pd.DataFrame.from_dict(d, orient="index", columns=notes_list)
        .reset_index()
        .rename(columns={"index": "id_produto_cor"})
        .replace(np.nan, 0)
    )


def consolidate_votes(
    merged_df: pd.DataFrame, dados_animale: pd.DataFrame
) -> pd.DataFrame:
    """
        Merges the sales+notes dataset with statistical measures for nota and nota_preco.

        Args:
            merged_df (:class:`pandas.Dataframe`): Sales data with repeating product, each with nota and nota_preco.
            dados_animale (:class:`pandas.Dataframe`): Sales data with all products.
        Returns:
            :class:`pandas.Dataframe`: The resulting merged Dataframe.
    """

    labels = ["nota", "nota_preco"]
    dados_agg = (
        merged_df[["id_produto_cor"] + labels]
        .groupby("id_produto_cor")
        .agg([np.mean, np.std, len])
    )

    dados_agg.columns = list(map("".join, dados_agg.columns.values))
    dados_agg = dados_animale.merge(dados_agg, on="id_produto_cor")

    dados_agg = dados_agg.merge(count_unique_votes(merged_df), on="id_produto_cor")

    return dados_agg



import mysql.connector
from mysql.connector import Error
import os
import pandas as pd
import re
import numpy as np
import warnings

from typing import List, Tuple


def connect_and_query(query: str) -> pd.DataFrame:
    """
    Opens a connection to a SQL database and returns the output the Query as a pandas DataFrame
    It is important to note that the environment variables need to be set prior to execution. 

    Args:
        query (str): The query to be fetched at the database.

    Returns:
        (:class:`pandas.Dataframe`):: The output of the query in DataFrame format.
    """
    mySQLconnection = None
    # for the next part, it is important to set the environment variables
    try:
        print("Opening MySQL Connection")
        mySQLconnection = mysql.connector.connect(
            host=os.environ.get("HOST_PLM_DATABASE"),
            database=os.environ.get("PLM_DATABASE"),
            user=os.environ.get("USER_PLM_DATABASE"),
            password=os.environ.get("PASSWORD_PLM_DATABASE"),
        )
        print("Connection Opened, starting Fetch")

        output = pd.read_sql(query, con=mySQLconnection)

    except Error as e:
        print("Error while connecting to MySQL", e)
        return e

    finally:
        # closing database connection.
        if mySQLconnection:
            if mySQLconnection.is_connected():
                mySQLconnection.close()
                print("MySQL connection is closed")
                return output


def format_produto_animale(produto: str) -> str:
    """
    Formats the 'produto' column to follow a xx.xx.x+ standard, thus removing forbidden characters (letters, specials, etc).

    Args:
        produto (str): A single entry from the produto column.

    Returns:
        (str): The formatted string if it fits the pattern. If no match is found returns nothing (empty string)

    """
    m = re.match("(^[0-9]{2}\.[0-9]{2}\.[0-9]+)", produto)
    if m:
        return m.groups()[0]
    else:
        return produto


def calculate_giro(
    df: pd.DataFrame, columns: list = ["value", "qtde", "preco", "qtty"]
) -> pd.DataFrame:
    """
    Calculates giro and disc from a dataframe

    Args:
        df (:class:`pandas.Dataframe`):: A correctly structured DataFrame
        columns (list(str)): The column names that represent:
            - value
            - qtde
            - preco
            - qtty
    
    Returns:
        (:class:`pandas.Dataframe`):: The original DataFrame with two additional columns,
        giro and discount.
    
    Todo:
        Complete the Args documentation.
    """
    # columns[0] = value, columns[1] = qtde, columns[2] = preco_varejo_original, columns[3] = qtty
    nan_indexes = []

    # checking if any important row has a zero (dividend row)
    for c in columns:
        if c == columns[0]:
            pass
        else:
            if df[c].eq(0).any().sum() > 0:
                column_indexes = np.where(df[columns[1]].eq(0).tolist())[0].tolist()
                nan_indexes = nan_indexes + column_indexes
                warnings.warn(
                    "{} column has at least one 0, resulting in NaN.".format(c)
                )

    warnings.warn("The NaN indexes are: {}".format(np.unique(nan_indexes)))

    # calculating giro
    df["giro"] = df[columns[0]] / (df[columns[1]] * df[columns[2]])
    # calculating discount
    df["disc"] = 1 - df[columns[0]] / (df[columns[3]] * df[columns[2]])

    return df


def clean_animale_dataset(df: pd.DataFrame) -> pd.DataFrame:
    """
        Cleans the animale sales dataset.

        It excludes infinite, zero quantities and NaN entries from the DataFrame.

        Args:
            df (:class:`pandas.Dataframe`): The animale sales dataset.

        Returns:
           :class:`pandas.Dataframe`: The cleaned dataset.
    """
    df.drop_duplicates(subset="id_produto_cor", keep=False, inplace=True)
    df["qtde"].replace(0, np.nan, inplace=True)
    df.replace([np.inf, -np.inf], np.nan, inplace=True)
    df.dropna(inplace=True)
    df.reset_index(drop=True, inplace=True)
    return df


def load_provao_dataset(
    path_animale: str, path_provao: str
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
        Loads the animale sales dataset (with sales indicators) and the votes dataset.

        Args:
            path_animale (str): Path to the animale dataset in Pickle format.
            path_provao (str): Path to the provao dataset in CSV format.
        
        Returns:
            (tuple): Tuple contaning:
                - df_merged (:class:`pandas.Dataframe`): A merged dataset from animale and provao.
                - df_animale (:class:`pandas.Dataframe`): Animale sales dataset.
                - df_provao (:class:`pandas.Dataframe`): Provao dataset      
    """

    df_animale = pd.read_pickle(path_animale)
    df_provao = pd.read_csv(path_provao)

    df_provao = df_provao.rename(columns={"preco": "nota_preco"})
    df_provao.drop(columns=["id_colecao", "id_produto_estilo", "produto"], inplace=True)

    df_animale = clean_animale_dataset(df_animale)
    df_animale.rename(columns={"preco_varejo_original": "preco"}, inplace=True)

    df_merged = df_animale.merge(df_provao, on="id_produto_cor")
    return df_merged, df_animale, df_provao


def reorganize_tags(tags: pd.DataFrame) -> pd.DataFrame:
    """
        Fixes the unorganized tags from the database, separating by 
        category.
    
        The value from the category is determined by the first uppercase
        letter.

        Args:
            tags (:class:`pandas.Dataframe`): 
        Returns:
            :class:`pandas.Dataframe`: The tags organized by category.

    """
    output = tags.copy
    output = tags.loc[:, []]
    for col in tags:
        for cat in tags[col].unique():
            if cat:
                aux = tags[tags[col] == cat][col]
                new_col = re.findall("([a-z]*)[A-Z]", cat)
                # Category value starts at the first uppercase letter.
                category_value = re.findall("([A-Z].+)", cat)
                if new_col:
                    new_col = new_col[0]
                    if new_col not in output.columns:
                        output[new_col] = None
                    output.loc[aux.index, new_col] = category_value
    return output


def label_encode_col(col):
    """
        Encodes a :class:`pandas.Series` with a label encoder that 
        accepts missing values (codified as a "None" string)

        Args:
            col (:class:`pandas.Series`): A column from a dataframe that 
            can include a missing entry codified as "None".
        Returns:
            :class:`pandas.Series`: The label-encoded column, with "None" 
            codified with label 0.
    """

    unique_values = col.unique()
    if "None" not in unique_values:
        unique_values = np.insert(unique_values, 0, "None")
    mapping_dict = {key: value for (value, key) in enumerate(unique_values)}
    reverse_dict = {key: value for (key, value) in enumerate(unique_values)}

    if "None" in mapping_dict:
        none_class = mapping_dict["None"]
        zero_key = reverse_dict[0]
        mapping_dict[zero_key], mapping_dict["None"] = (none_class, 0)

    # print(mapping_dict)
    return col.apply(lambda x: mapping_dict[x])



#%%
from soma.estilo import modelling, utils
import pandas as pd
from sklearn.preprocessing import LabelEncoder
import pickle
import os

class Datasets:
    def __init__(self):
        self.train = None
        self.test = None
        self.path = None
        self.desc = None
        self.label_columns = None
        self.dataset = None

    def load_sales(
        self,
        encoder=None,
        train_collections=["VER18", "INV18"],
        test_collections=["VER19"],
        clf=False,
    ):

        try:
            with open(os.environ.get('SALES_DATASET_PATH'), "rb") as f:
                dataset = pd.read_pickle(f)
        except:
            dataset = self.create_sales()

        col = [
            "qtde",
            "linha",
            "grupo",
            "preco",
            "disc",
            "category",
            "apparel",
            "productpatern",
            "product",
            "bodygarment",
            "giro",
        ]

        train_mask = dataset["colecao"].isin(train_collections)
        test_mask = dataset["colecao"].isin(test_collections)
        
        self.dataset = dataset

        if clf:
            top_giro = dataset[train_mask]['giro'].mean() + dataset[train_mask]['giro'].std()*0.5
            bot_giro = dataset[train_mask]['giro'].mean() - dataset[train_mask]['giro'].std()*0.5
            dataset['giro'] = dataset['giro'].apply(lambda x: 1*(x > top_giro) + 2*(x < bot_giro))
            self.type = 'CLF'
        else:
            self.type = 'REG'

        dataset = dataset[col]

        if encoder:
            encoder = encoder(cols=list(set(self.label_columns) & set(col)))
            dataset = encoder.fit_transform(dataset)
            self.desc = 'Encoded_dataset'

        dataset_train = dataset[train_mask][:]
        dataset_test = dataset[test_mask][:]

        self.train = (dataset_train.to_numpy()[:,:-1], dataset_train.to_numpy()[:,-1])
        self.test = (dataset_test.to_numpy()[:,:-1], dataset_test.to_numpy()[:,-1])
        self.desc = 'Labelled_dataset'

        return self

    def create_sales(self):
        path_animale = os.environ.get('ANIMALE_SALES_PATH')
        path_tags = os.environ.get('ANIMALE_TAGS_PATH')

        try:
            fixed_tags = pd.read_pickle(path_tags)
        except:
            with open("tags_query.sql", "r") as f:
                query = f.read()
            tags = utils.connect_and_query(query)

            tags = tags.groupby("id_produto_cor").apply(
                lambda x: x["desc_Fashion_Attributes"].str.cat(sep=",")
            )
            tags = pd.DataFrame(tags).reset_index()
            tags.columns = ["id", "nome"]
            tags = tags.set_index("id")
            split_tags = tags["nome"].str.split(",", expand=True)
            split_tags = split_tags.set_index(tags.index.values)

            fixed_tags = utils.reorganize_tags(split_tags).reset_index()
            fixed_tags.rename(
                columns={fixed_tags.columns[0]: "id_produto_cor"}, inplace=True
            )

            with open(path_tags, "wb") as f:
                pickle.dump(fixed_tags, f)

        try:
            animale_df = pd.read_pickle(path_animale)
        except:
            print("File Doesn't exist")
            return 0

        animale_df = utils.clean_animale_dataset(animale_df)
        animale_df.rename(columns={"preco_varejo_original": "preco"}, inplace=True)
        animale_df.drop(
            animale_df.loc[animale_df["grupo"] == "SAPATOS"].index, inplace=True
        )
        animale_df.drop(
            animale_df.loc[animale_df["grupo"] == "SAPATILHA"].index, inplace=True
        )
        animale_df.drop(
            animale_df.loc[animale_df["grupo"] == "BOLSAS"].index, inplace=True
        )
        animale_df.drop(
            animale_df.loc[animale_df["grupo"] == "DIVERSOS"].index, inplace=True
        )
        animale_df.drop(
            animale_df.loc[animale_df["grupo"] == "CINTOS"].index, inplace=True
        )
        animale_df.drop(
            animale_df.loc[animale_df["grupo"] == "ACESSORIOS"].index, inplace=True
        )

        animale_df = animale_df.merge(fixed_tags, on="id_produto_cor")
        animale_df.reset_index(inplace=True, drop=True)
        animale_df.drop(
            [
                "bag",
                "clutch",
                "sneakers",
                "sandal",
                "boots",
                "sandals",
                "heel",
                "footwear",
                "shoe",
                "shoes",
                "toe",
            ],
            axis=1,
            inplace=True,
        )

        animale_df["grupo"].replace(r"^(TOP) (.*)", r"\1", regex=True, inplace=True)
        animale_df["grupo"].replace(r"^(OVERTOP)(.*)", r"\1", regex=True, inplace=True)

        animale_df["bodygarment"] = (
            animale_df["lowerbodygarment"].astype(str)
            + animale_df["upperbodygarment"].astype(str)
            + animale_df["fullbodygarment"].astype(str)
        )
        animale_df["bodygarment"].replace(
            r"^None(.*)None", r"\1", regex=True, inplace=True
        )
        animale_df.drop(
            columns=["lowerbodygarment", "upperbodygarment", "fullbodygarment"],
            inplace=True,
        )
        animale_df[animale_df.columns[11:]] = animale_df[
            animale_df.columns[11:]
        ].astype(str)

        self.label_columns = ["linha", "grupo"] + animale_df.columns[11:].to_list()

        for col in self.label_columns:
            animale_df[col] = utils.label_encode_col(animale_df[col])

        with open(os.environ.get('SALES_DATASET_PATH'), "wb") as f:
            pickle.dump(animale_df, f)

        return animale_df

